/* Pawe Kapica H1ISI
 * 
 * Wzorzec projektowy kompozycji
 * 
 * 1. Zastosowanie:
 * Reprezentacja drzewiastych struktur danych.
 * W drzewiastej strukturze danych wyrniamy:
 * 	- klasy proste (bdce liciami w drzewie),
 *	- klasy zoone realizujce agregacj klas prostych i zoonych (bdce wzami wewntrznymi w drzewie lub liciami).
 * 
 * Istnieje cz wsplna funkcjonalnoci wszystkich klas prostych i zoonych,
 * klasy zoone realizuj j zwykle przez wywoanie danej funkcji dla wszystkich swoich dzieci.
 * Klasy zoone posiadaj dodatkowo funkcjonalno zwizan z dodawaniem i usuwaniem dzieci.
 * 
 * 2. Idea:
 * Stworzenie wsplnej klasy bazowej (komponent) stanowicej interfejs dla klas prostych i zoonych.
 * Klasa zoona jest komponentem, moe agregowa komponenty wszystkich innych typw (zarwno klas prostych jak i zoonych)
 * 
 * 3. Decyzja projektowa:
 * Czy klasa bazowa (komponent) powinna deklarowa funkcjonalno specyficzn klas zoonych (dodawanie, usuwanie dzieci).
 * Uniwersalno vs. bezpieczestwo typw.
 * 
 * 4. Zalety:
 * Pozwala klientowi traktowa jednakowo wszystkie klasy proste i zoone (wsply interfejs komponentu).
 * Manipulowanie grup obiektw takie same jak prostym obiektem
 * (wsplny kod manipulacji obiektami, zwize rekurencyjne wywoania funkcji wewntrz struktury).
 * atwo dodawania nowych klas do hierarchii.
 * 
 * 5. Powszechnie stosowana wsppraca z innymi wzorcami, m.in.:
 * 	- Kompozycja - dekorator
 * 		Decyzja projektowa: komponent deklaruje funkcjonalno klas zoonych - uniwersalno (ewentualnie wiele dekoratorw)
 * 
 * 	- Kompozycja - pyek (flyweight)
 * 		Komponenty nie posiadaj wskanika na rodzica (ewentualnie posiadaj list wskanikw na wszystkich rodzicw)
 * 
 * 	- Kompozycja - iterator - wizytator
 * 		Iterowanie po rekurencyjnej strukturze drzewa bez wiedzy o jej fizycznej strukturze, wizytacja komponentw
 * 
 * 	- Kompozycja - strategia - obserwator
 */
 
/*
 * Przykad: system plikw: katalogi i pliki
 */
#include <iostream>
#include <cstdlib>
#include <string>
#include <vector>

using namespace std;

/* Abstrakcyjna klasa bazowa reprezentujca komponent */
class AbstractFile{
	public:
		/* Funkcja wsplna klas prostych i zoonych*/
		virtual void list() const = 0;
	
		virtual ~AbstractFile(){}
};

/* Klasa prosta, reprezentuje pojedynczy plik */
class File : public AbstractFile{
	public:
		File(string name) : name_(name){}

		/* Wypisanie wasnego identyfikatora */
		void list() const {cout << name_ << endl;}

	private:
		const string name_;
};

/* Klasa zoona, reprezentuje katalog */
class Directory : public AbstractFile{
	public:
		Directory(const string& name) : name_(name){}
		
		void list() const { 
			/* wypisanie wasnego identyfikatora...*/
			cout << name_ << endl;

			/* ... oraz identyfikatory wszystkich dzieci */
			for(vector<AbstractFile*>::const_iterator it = content_.begin(); it != content_.end(); ++it)
				(*it)->list();
		}
		
		void add(AbstractFile* file){ content_.push_back(file); }
		
	private:
		vector<AbstractFile*> content_;
		const string name_;
};

/* prosta hierarchia katalogw i plikw */
int main(int argc, char **argv) {
	Directory* root = new Directory("/");
	Directory* etc = new Directory("etc");
	Directory* home = new Directory("home");
	home->add(new Directory("jump"));
	root->add(home);
	etc->add(new File("fstab")); 
	etc->add(new File("passwd"));
	root->add(etc);
	
	/* rekurencyjne wylistowanie drzewa komponentw*/
	root->list();
	
	return EXIT_SUCCESS;	
}

/* Wynik na standardowy wyjciu:
 * 
 * home
 * jump
 * etc
 * fstab
 * passwd
 */
